print("import: 'boltons'")
import boltons

print("import: 'boltons.cacheutils'")
import boltons.cacheutils

print("import: 'boltons.debugutils'")
import boltons.debugutils

print("import: 'boltons.dictutils'")
import boltons.dictutils

print("import: 'boltons.ecoutils'")
import boltons.ecoutils

print("import: 'boltons.fileutils'")
import boltons.fileutils

print("import: 'boltons.formatutils'")
import boltons.formatutils

print("import: 'boltons.funcutils'")
import boltons.funcutils

print("import: 'boltons.gcutils'")
import boltons.gcutils

print("import: 'boltons.ioutils'")
import boltons.ioutils

print("import: 'boltons.iterutils'")
import boltons.iterutils

print("import: 'boltons.jsonutils'")
import boltons.jsonutils

print("import: 'boltons.listutils'")
import boltons.listutils

print("import: 'boltons.mathutils'")
import boltons.mathutils

print("import: 'boltons.mboxutils'")
import boltons.mboxutils

print("import: 'boltons.namedutils'")
import boltons.namedutils

print("import: 'boltons.pathutils'")
import boltons.pathutils

print("import: 'boltons.queueutils'")
import boltons.queueutils

print("import: 'boltons.setutils'")
import boltons.setutils

print("import: 'boltons.socketutils'")
import boltons.socketutils

print("import: 'boltons.statsutils'")
import boltons.statsutils

print("import: 'boltons.strutils'")
import boltons.strutils

print("import: 'boltons.tableutils'")
import boltons.tableutils

print("import: 'boltons.tbutils'")
import boltons.tbutils

print("import: 'boltons.timeutils'")
import boltons.timeutils

print("import: 'boltons.typeutils'")
import boltons.typeutils

print("import: 'boltons.urlutils'")
import boltons.urlutils

